#
# The sh module, which is the main ingredient of pysh.
#
# Copyright 2003-2004 by Max Neunhoeffer
#
# $Id: sh.py,v 1.12 2005/04/26 13:58:15 neunhoef Exp $
#

import sys,os,string,time,re,signal,pwd,traceback,socket,threading,resource
import termios,signal
from glob import glob
from sys import stdin,stdout,stderr

class ShError(Exception):
    def __init__(self,value):
        self.value = value
    def __str__(self):
        return `self.value`

WeDoJobControl = 0     # set to 1 if sh should do job control
TermSettings = None    # here we store our terminal settings

dirstack = []

if os.environ.has_key("PWD"):
    pwd = os.environ['PWD']
    if os.path.realpath(pwd) != os.getcwd():
        pwd = os.getcwd()
        os.environ['PWD'] = pwd
    dirstack.append(pwd)
else:
    pwd = os.getcwd()
    dirstack.append(pwd)
    os.environ['PWD'] = pwd
    
# Determine home directory:
homedir = ''
if os.environ.has_key('USER'):
    try:
        pwent = pwd.getpwnam(os.environ['USER'])
        homedir = pwent[5]
    except:
        pass
if homedir == '' and os.environ.has_key('HOME'):
    homedir = os.environ['HOME']

# The following dictionary contains helper functions to configure
# the completion algorithm:
completerprogs = {}

# Aliases:
aliases = {}

# History:
history = []
historyfile = os.path.join(homedir,'.pyshhist')

# process management:
children = {}   # a dictionary with an entry for each pid of a child
                # all entries are pairs of a job and a number in there

path = []     # a list of paths
pcache = []   # a list, each entry is either None or a dictionary with
              # keys being the filenames and values being 1
cachelock = threading.Lock()     # lock for access to pcache
rehashthread = None              # here we store a thread for rehashing
                                 # during its runtime
rehashevent = threading.Event()  # an event to synchronize the rehash thread
rehashdone = threading.Event()   # indicates that rehash is ready

def init_hashing():
    '''This function should be called to (re-)initialize the path lookup.'''
    global path,pcache,cachelock
    cachelock.acquire()
    path = os.environ['PATH'].split(':')
    pcache = map(lambda x: None,path)
    # clean paths:
    i = 0
    while i < len(path):
        if len(path[i]) == 0:
            del path[i]
            del pcache[i]
        else:
            if path[i][0:2] == '~/': 
                path[i]=os.path.join(homedir,path[i][2:])
            i += 1
    cachelock.release()

def rehash_thread_func():
    """This is called in a daemonic subthread and does the real work."""
    global path,pcache,rehashevent,rehashdone

    while 1:   # repeat infinitely
        rehashdone.clear()   # not yet done
        rehashevent.clear()  # we start rehashing
        init_hashing()
        # now we work on the path:
        i = 0
        while i < len(path):
            if path[i][0] == '.':
                i += 1
            elif os.path.isdir(path[i]):
                l = os.listdir(path[i])
                d = {}
                for f in l: 
                    p = os.path.join(path[i],f)
                    if os.access(p,os.X_OK) and not(os.path.isdir(p)): d[f] = 1
                cachelock.acquire()
                pcache[i] = d
                cachelock.release()
                i += 1
            else:
                cachelock.acquire()
                pcache[i] = {}
                cachelock.release()
                i += 1
        if not(rehashevent.isSet()):
            # If this is set, we already have to do another one:
            rehashdone.set()    # indicate that rehash is done
            rehashevent.wait()  # wait for event


def rehash(wait = 0):
    # This function starts the rehash thread or, if this is already done,
    # sends a rehash-event to the other thread.
    # If wait is true, then this function returns only after the rehash
    # is done. This can take a while for NFS file systems.
    global rehashthread,rehashevent,rehashdone

    # first wait for old thread to clear the cache:
    if rehashthread == None:
        # now start the subthread as a daemon:
        rehashevent = threading.Event()
        rehashevent.clear()
        rehashthread = threading.Thread(target=rehash_thread_func)
        rehashthread.setDaemon(1)
        rehashthread.start()
    else:
        rehashevent.set()   # the other thread will take care of everything!
    if wait:
        rehashdone.wait()

def which(f):
    if f.find(os.sep) >= 0: 
        if os.access(f,os.X_OK) and not(os.path.isdir(f)): return f
        else: return None
    i = 0
    while i < len(path):
        cachelock.acquire()
        if path[i][0] == '.' or pcache[i] == None:
            cachelock.release()
            p = os.path.join(path[i],f)
            if os.access(p,os.X_OK) and not(os.path.isdir(p)): return p
        elif pcache[i].has_key(f):
            cachelock.release()
            return os.path.join(path[i],f)
        else:
            cachelock.release()
        i += 1
    return None
        
def builtin_which(args):
    '''Builtin shell command "which".'''
    g = globals()
    for a in args[1:]:
        if g.has_key('builtin_'+a):
            sys.stdout.write(a+': shell builtin'+'\n')
            return
        r = which(a)
        if r != None:
            sys.stdout.write(r+'\n')
        else:
            sys.stdout.write(a+" not found\n")

def builtin_rehash(args):
    rehash()

def builtin_cd(args):
    global dirstack
    if len(args) == 1:
        d = homedir
    else:
        d = args[1]
        if d[0] != '/':
            d = os.path.normpath(os.path.join(dirstack[-1],d))
    
    if d != '':
        try:
            os.chdir(d)
            if os.path.realpath(d) != os.getcwd():
                d = os.getcwd()
            dirstack.append(d)
            if len(dirstack) > 100:
                dirstack = dirstack[50:]
            os.environ['PWD'] = d
        except:
            sys.stderr.write('Error: no such directory\n')
            return

def unexpandtilde(d):
    if d[:len(homedir)] == homedir: return '~'+d[len(homedir):]
    else: return d

def builtin_pwd(args):
    sys.stdout.write(unexpandtilde(dirstack[-1])+'\n')

def builtin_popd(args):
    if len(dirstack) <= 1:
        sys.stderr.write("Error: directory stack empty\n")
        return
    del dirstack[-1]
    d = dirstack.pop()
    builtin_cd(['cd',d])
    builtin_pwd(['pwd'])

def builtin_dirs(args):
    for d in dirstack: 
        if d[:len(homedir)] == homedir: 
            sys.stdout.write('~'+d[len(homedir):]+' ')
        else:
            sys.stdout.write(d+' ')
    sys.stdout.write('\n')

def builtin_jobs(args):
    update_jobinfo()
    l = job.jobs.keys()
    l.sort()
    for i in l:
        j = job.jobs[i]
        sys.stdout.write('Job #'+str(i)+' ')
        for k in range(len(j.args)):
            sys.stdout.write(j.args[k][0]+' (')
            if j.termed[k]:
                sys.stdout.write('terminated')
            else:
                sys.stdout.write(str(j.pids[k]))
                if j.stopped[k]: sys.stdout.write("T")
            sys.stdout.write(') ')
            if k < len(j.args)-1: sys.stdout.write('| ')
        sys.stdout.write('\n')
        
def builtin_exit(args):
    sys.exit(0)

namere = re.compile('^[a-zA-Z_0-9]+$')

def builtin_setenv(args):
    if len(args) < 2:
        sys.stderr.write('setenv: bad assignment\n')
        raise ShError, 'setenv: bad assignment'
    k = args[1]
    if not(namere.match(k)):
        sys.stderr.write('setenv: illegal variable name: '+k+'\n')
        raise ShError, 'setenv: illegal variable name'
    if len(args) == 2:
        if os.environ.has_key(k): del os.environ[k]
        return
    v = args[2]
    os.environ[k] = v

builtin_unsetenv = builtin_setenv
builtin_set = builtin_setenv
builtin_unset = builtin_setenv

assignre = re.compile('^([a-zA-Z_0-9]+)=(.*)$')

def builtin_export(args):
    if len(args) < 2:
        sys.stderr.write('export: no assignment\n')
        raise ShError, 'export: no assignment\n'
    for a in args[1:]:
        m = assignre.match(a)
        if not(m):
            m = namere.match(a)
            if not(m):
                sys.stderr.write('export: bad assignment: '+a+'\n')
                raise ShError, 'export: bad assignment: '+a
            return
        os.environ[m.group(1)] = m.group(2)
         
def builtin_bg(args):
    if len(args) <= 1:
        l = job.jobs.keys()
        if len(l) == 0:
            sys.stderr.write("bg: no current job\n")
            raise ShError, 'bg: no current job'
        l.sort()
        j = job.jobs[l[0]]
    else:
        try:
            i = int(args[1])
            j = job.jobs[i]
        except:
            sys.stderr.write("bg: unknown job: "+args[1]+"\n")
            raise ShError, 'bg: unknown job'
    # send the CONT signal (if jobs are not yet termed):
    try: os.kill(-j.pids[0],signal.SIGCONT)
    except: pass
    j.nrrunning = len(j.pids) - j.nrtermed
    j.stopped = [0] * len(j.pids)
    sys.stderr.write('Job #'+str(j.id)+' continued.\n')

def builtin_fg(args):
    if len(args) <= 1:
        l = job.jobs.keys()
        if len(l) == 0:
            sys.stderr.write("fg: no current job\n")
            raise ShError, 'fg: no current job'
        l.sort()
        j = job.jobs[l[0]]
    else:
        try:
            i = int(args[1])
            j = job.jobs[i]
        except:
            sys.stderr.write("fg: unknown job: "+args[1]+"\n")
            raise ShError, 'fg: unknown job'
    # send the CONT signal (if jobs are not yet termed):
    if WeDoJobControl:
        os.tcsetpgrp(sys.stdin.fileno(),j.pids[0])
        if j.term != None:
            termios.tcsetattr(sys.stdin.fileno(),termios.TCSADRAIN,j.term)
    try: os.kill(-j.pids[0],signal.SIGCONT)
    except: pass
    j.nrrunning = len(j.pids) - j.nrtermed
    j.stopped = [0] * len(j.pids)
    j.waitready()

def builtin_umask(args):
    if len(args) <= 1:
        u = os.umask(0)
        os.umask(u)
        us = oct(u)
        while len(us) < 3: us = '0'+us
        sys.stdout.write(us[-3:]+'\n')
    else:
        a = args[1]
        while len(a) < 3: a = '0'+a
        a = a[-3:]
        os.umask(64*(ord(a[0])-ord('0')) + 8*(ord(a[1])-ord('0')) +
                 ord(a[2]) - ord('0'))
        
def builtin_alias(args):
    global aliases
    if len(args) <= 1:
        l = aliases.keys()
        l.sort()
        for a in l:
            sys.stdout.write(a+" ==> "+string.join(aliases[a]," ")+'\n')
        return
    if len(args) <= 2:
        if not(aliases.has_key(args[1])):
            sys.stderr.write("alias: unknown alias: "+args[1]+"\n")
            raise ShError, 'alias: unknown alias'
        del aliases[args[1]]
    else:
        aliases[args[1]] = string.split(args[2]," ")

signal_table = {'HUP':1,'INT':2,'QUIT':3,'ILL':4,'TRAP':5,'ABRT':6,'BUS':7,
                'FPE':8,'KILL':9,'USR1':10,'SEGV':11,'USR2':12,'PIPE':13,
                'ALRM':14,'TERM':15,'STKFLT':16,'CHLD':17,'CONT':18,'STOP':19,
                'TSTP':20,'TTIN':21,'TTOU':22,'URG':23,'XCPU':24,'XFSZ':25,
                'VTALRM':26,'PROF':27,'WINCH':28,'POLL':29,'PWR':30,'SYS':31}

def builtin_kill(args):
    signal = 15
    for a in args[1:]:
        if a[0:1] == '-':   # change current signal
            try: signal = signal_table[a[1:]]
            except:
                try: signal = int(a[1:])
                except:
                    sys.stderr.write("Warning: Unknown signal: "+a[1:]+
                                     " assuming TERM.\n")
                    signal = 15
        elif a[0:1] == '%':   # one of our jobs:
            try: 
                jobnr = int(a[1:])
                j = job.jobs[jobnr]
            except:
                sys.stderr.write("Error: Unknown job: "+a[1:]+" ... ignored.\n")
                continue
            try: os.kill(-j.pids[0],signal)
            except: pass
        else:
            try:
                pid = int(a)
            except:
                sys.stderr.write('Error: No process ID: "'+a+'" ... ignored.\n')
                continue
            try: os.kill(pid,signal)
            except:
                sys.stderr.write("Error: No such process: "+str(pid)+".\n")

def builtin_savehist(args):
    name = historyfile
    if len(args) > 1:
        name = args[1]
    try:
        f = file(name,"w")
        for l in history[:-1]:
            f.write(l+"\n")
        f.close()
    except:
        sys.stderr.write('Error: Cannot write history to "'+name+'".\n')

def builtin_loadhist(args):
    global history
    name = historyfile
    if len(args) > 1:
        name = args[1]
    try:
        f = file(name)
        history = map(string.rstrip,f.readlines())
        f.close()
    except:
        sys.stderr.write('Error: Cannot load history from "'+name+'".\n')
    
limittab = {'cputime':resource.RLIMIT_CPU,'filesize':resource.RLIMIT_FSIZE,
            'datasize':resource.RLIMIT_DATA,'stacksize':resource.RLIMIT_STACK,
            'coredumpsize':resource.RLIMIT_CORE,
            'memoryuse':resource.RLIMIT_RSS,'maxproc':resource.RLIMIT_NPROC,
            'descriptors':resource.RLIMIT_NOFILE,
            'memorylocked':resource.RLIMIT_MEMLOCK,
            'addressspace':resource.RLIMIT_AS}

def builtin_limit(args):
    hard = 0
    if len(args) == 1:   # Show limits:
        sys.stdout.write("%-13s%12s%12s\n" % ("Limit","Soft","Hard"))
        ll = limittab.keys()
        ll.sort()
        for l in ll:
            (s,h) = resource.getrlimit(limittab[l])
            sys.stdout.write("%-13s%10d%10d\n" % (l,s,h))
    else:
        if len(args) == 4 and args[1] == '-h':
            hard = 1
            res = args[2]
            lim = args[3]
        elif len(args) == 3:
            res = args[1]
            lim = args[2]
        else:
            sys.stderr.write("Usage: limit [-h] RESOURCE LIMIT\n")
            raise ShError, "Usage: limit"
        if not(limittab.has_key(res)):
            sys.stderr.write("Error: Unknown resource: "+args[i]+".\n")
            raise ShError, "Error: Unknown resource"
        try:
            if lim[0:1] == 'u':
                lim = -1
            elif lim[-1:] == 'k':
                lim = int(lim[:-1])*1024
            elif lim[-1:] == 'm':
                lim = int(lim[:-1])*1024*1024
            elif lim[-1:] == 'h':
                lim = int(lim[:-1])*3600
            else:
                lim = int(lim)
        except:
            lim = -1
        try:
            (s,h) = resource.getrlimit(limittab[res])
            if hard:
                resource.setrlimit(limittab[res],(s,lim))
            else:
                resource.setrlimit(limittab[res],(lim,h))
        except:
            sys.stderr.write('Error: Cannot set ')
            if hard: sys.stderr.write('hard')
            else: sys.stderr.write('soft')
            sys.stderr.write(' resource '+res+' to '+str(lim)+' .\n')
            raise ShError, "Error: cannot set resource."
            

# Our own improvement of globbing taking care of "{}" pairs:
def doglobbing(s):
    """Takes care of braces, returns list of strings."""
    #if s[0:2] == '~/':
    #    s = os.path.join(homedir,s[2:])
    b = s.find('{')
    e = s.find('}',b)
    p = s.find(',',b)
    if b >= 0 and p > b and e > p:    #   {*,*}
        result = []
        for t in s[b+1:e].split(','): 
            result.extend(doglobbing(s[:b]+t+s[e+1:]))
        return result
    # now try globbing:
    if s.find('*') >= 0 or s.find('?') >= 0 or \
       (s.find('[') >= 0 and s.find(']') >= 0):
        return glob(s)
    else:
        return [s]

def onlywithext(l,ext):
    i = 0
    ll = len(ext)
    exts = ext+' '
    while i < len(l):
        if l[i][-1:] == '/' or l[i][-ll:] == ext or l[i][-ll-1:] == exts: i+=1
        else: del(l[i])

def onlydirs(l,ext):
    i = 0
    while i < len(l):
        if l[i][-1:] == '/': i += 1
        else: del(l[i])

gotsuspend = 0
def suspend_handler(s,f):
    global gotsuspend
    gotsuspend = 1

lastexitcode = 0

class job:
    """Stands for one pipe of processes."""
    jobs = {}       # this is a class variable where all jobs are stored
    id = 0          # job id's in shell
    pids = []       # process ids, [] means: not yet launched
    args = []       # a list of lists of strings, each entry corresponds to
                    # a process in the pipe args[i][0] is the command
    doglob = []     # like args, but for every argument 0 or 1, indicating
                    # whether globbing is done
    background = 0  # flag, whether job is in the background
    termed = []     # flag, whether process is terminated
    stopped = []    # flag, whether process is stopped
    nrrunning = 0   # number of processes not stopped and not terminated
    exitcode = []   # exit code, only counts, when done
    collector = []  # list to collect characters for next argument
    collglob = 0    # here we note, whether we will do globbing
    input = None    # file descriptor that will be dupped to stdin in child
    output = None   # file descriptor that will be dupped to stdout in child
    error = []      # file descriptors that will be dupped to stderr
    redirect = ''   # current mode of redirection, only for parsing
    toclose = []    # list of file descriptors to close in children
    term = None     # terminal settings if job was stopped

    def __init__(self,background = 0):
        i = 1
        while self.jobs.has_key(i): i += 1
        self.id = i
        self.jobs[i] = self
        self.pids = [0]
        self.args = [[]]
        self.termed = [0]
        self.stopped = [0]
        self.nrrunning = 0
        self.nrtermed = 0
        self.exitcode = [0]
        self.background = background
        self.collector = []
        self.input = None
        self.output = None
        self.error = [None]
        self.redirect = ''
        self.toclose = []

    def doglob(self,s):
        self.args[-1].extend(doglobbing(s))

    def appendarg(self):
        if len(self.collector) > 0:
            if self.collglob:
                self.doglob(string.join(self.collector,''))
            else:
                self.args[-1].append(string.join(self.collector,''))
            self.collector = []
            self.collglob = 0

    def reportstopped(self):
        sys.stderr.write('Job #'+str(self.id)+' stopped.\n')
        
    def reportterminated(self):
        if not(self.background):
            doreport = 0
            for i in self.exitcode:
                if i != 0: 
                    doreport = 1
            if doreport == 0: return
        m = 'Job #'+str(self.id)+' completed'
        try:
            m += ': '+str(self.exitcode)+'\n'
        except:
            m += '\n'
        sys.stderr.write(m)

    def addprocess(self):
        self.appendarg()
        self.pids.append(0)
        self.args.append([])
        self.termed.append(0)
        self.stopped.append(0)
        self.exitcode.append(0)
        self.error.append(None)

    def waitready(self):
        '''Wait for a job to terminate or stop (all processes).'''
        global TermSettings
        while self.nrrunning > 0:
            update_jobinfo(canblock = 1)
        if WeDoJobControl:
            if self.nrtermed < len(self.pids):   # stopped, not all are termed:
                self.term = termios.tcgetattr(sys.stdin.fileno())
            os.tcsetpgrp(sys.stdin.fileno(),os.getpgrp())
            termios.tcsetattr(sys.stdin.fileno(),termios.TCSADRAIN,TermSettings)

    def launch(self):
        '''Launches processes. If an error occurs, an exception is raised.
        if self.background is non-zero, return immediately after launching.
        Otherwise wait until job has completed. self.input and self.output are
        None or file descriptors, which must be open file descriptors.
        The file descriptors will be dupped to 0 and 1 in the children
        and will be closed in the parent by this call!
        self.error is a list. Each entry must be a file descriptors or None.
        The length must be equal to the number of processes
        in the pipe. A file descriptor in a position corresponding
        to a job will be dupped to stderr in that child and will be closed
        in the parent.  The file descriptors in self.toclose will be closed 
        in the children.'''
        global TermSettings
        # allow the empty command:
        if len(self.args) == 1 and len(self.args[-1]) == 0:
            self.cleanup()
            return
        
        nrproc = len(self.args)

        # Replace aliases:
        for i in range(nrproc):
            if len(self.args[i]) > 0 and aliases.has_key(self.args[i][0]):
                newargs = []
                for st in aliases[self.args[i][0]]:
                    s = st
                    pos = 0
                    while 1:
                        pos = s.find("%",pos)
                        if pos < 0: break
                        c = s[pos+1:pos+2]
                        if c == '%':
                            s = s[:pos+1] + s[pos+2:]
                            pos += 1
                        elif c >= '1' and c <= '9':
                            nr = ord(c) - ord('0')
                            if len(self.args[i]) > nr:
                                s = s[:pos] + self.args[i][nr] + s[pos+2:]
                                pos += len(self.args[i][nr])
                            else:
                                s = s[:pos] + s[pos+2:]
                        else:
                            pos += 1
                    newargs.append(s)
                self.args[i][0:1] = newargs

        # check for shell builtins:
        if len(self.args[0]) > 0:
             com = 'builtin_'+self.args[0][0]
             g = globals()
             if g.has_key(com):
                 # Input/Output redirection:
                 if self.input != None:
                     savein = os.dup(0)
                     os.dup2(self.input,0)
                 if self.output != None:
                     saveout = os.dup(1)
                     os.dup2(self.output,1)
                 self.cleanup()
                 g[com](self.args[0])
                 if self.input:
                     os.dup2(savein,0)
                 if self.output:
                     os.dup2(saveout,1)
                 return
        
        lastpipe = None
        progs = []
        for i in range(nrproc):
            if len(self.args[i]) == 0: 
                sys.stderr.write('Error: empty command in pipe\n')
                self.cleanup()
                raise ShError,"empty command in pipe"
            progs.append(which(self.args[i][0]))
            if progs[-1] == None:
                sys.stderr.write('Error: command not found: '+
                                 self.args[i][0]+'\n')
                self.cleanup()
                raise ShError,"command not found: "+self.args[i][0]
        
        # Save our terminal settings:
        if WeDoJobControl:
            TermSettings = termios.tcgetattr(sys.stdin.fileno())

        pgid = 0
        for i in range(nrproc):
            if i < nrproc-1: 
                try:
                    newpipe = os.pipe()
                except:
                    sys.stderr.write('Error: Cannot create pipe!\n')
                    if lastpipe != None:
                        os.close(lastpipe[0])
                        os.close(lastpipe[1])
                    raise ShError, 'no pipe'
            else: newpipe = None
            self.pids[i] = os.fork()
            if self.pids[i] == 0:   # the child
                if WeDoJobControl:
                    pid = os.getpid()
                    if i == 0: pgid = pid
                    os.setpgid(pid,pgid)
                    if not(self.background):
                        os.tcsetpgrp(sys.stdin.fileno(),pgid)
                    signal.signal(signal.SIGINT,  signal.SIG_DFL)
                    signal.signal(signal.SIGQUIT, signal.SIG_DFL)
                    signal.signal(signal.SIGTSTP, signal.SIG_DFL)
                    signal.signal(signal.SIGTTIN, signal.SIG_DFL)
                    signal.signal(signal.SIGTTOU, signal.SIG_DFL)
                    signal.signal(signal.SIGCHLD, signal.SIG_DFL)
                # switch stdin and stdout
                if i > 0:   # stdin, lastpipe is set!
                    os.dup2(lastpipe[0],0)
                    os.close(lastpipe[0])
                    os.close(lastpipe[1])
                else:       # stdin redirected?
                    if self.input != None:
                        os.dup2(self.input,0)
                        os.close(self.input)
                if i < nrproc-1:   # stdout, newpipe is set!
                    os.dup2(newpipe[1],1)
                    os.close(newpipe[0])
                    os.close(newpipe[1])
                else:       # stdout redirected?
                    if self.output != None:
                        os.dup2(self.output,1)
                        os.close(self.output)
                # switch err:
                if len(self.error) > i and self.error[i] != None:
                    os.dup2(self.error[i],2)
                    if self.error[i] != 1: 
                        os.close(self.error[i])
                        self.error[i] = None
                for d in self.toclose:
                    os.close(d)
                try:
                    os.execv(progs[i],self.args[i])
                except:
                    os._exit(1)
            # the parent:
            # note new process:
            self.nrrunning += 1
            if WeDoJobControl:
                # To avoid race conditions we do this here and in the child:
                if i == 0: pgid = self.pids[0]
                try: os.setpgid(self.pids[i],pgid)
                except: pass    # ignore error, if child has already done this
            children[self.pids[i]] = (self,i)   # remember child
            if lastpipe:
                os.close(lastpipe[0])
                os.close(lastpipe[1])
            lastpipe = newpipe
            if len(self.error) > i and self.error[i] != None and \
               self.error[i] != 1: 
                   os.close(self.error[i])
                   self.error[i] = None
        # Set controling terminal
        if WeDoJobControl and not(self.background):
            os.tcsetpgrp(sys.stdin.fileno(),pgid)
        if self.input != None: 
            os.close(self.input)
            self.input = None
        if self.output != None: 
            os.close(self.output)
            self.output = None
        if not(self.background):   # wait for processes:
            self.waitready()
        else:
            sys.stderr.write('Job #'+str(self.id)+' PIDs='+str(self.pids)+
                             ' launched.\n')

    def cleanup(self):
        '''Cleanup everything when something went horribly wrong.'''
        if self.input != None: os.close(self.input)
        if self.output != None: os.close(self.output)
        for i in self.error:
            if i != None and i != 1: os.close(i)
        del self.jobs[self.id]

def update_jobinfo(canblock = 0):
    '''Fetches new information about jobs/processes via waitpid.'''
    global children
    while 1:
        try:
          (pid,status) = os.waitpid(-1,
                              ((1-canblock) * os.WNOHANG) | os.WUNTRACED)
        except OSError:
          if canblock == 0: break   # obviously no job to wait for
          # otherwise something fishy is going on:
          traceback.print_exc()
          pid = 0
          status = 0
        except:
          traceback.print_exc()
          pid = 0
          status = 0
        if children.has_key(pid):
            (j,i) = children[pid]
            j.nrrunning -= 1
            if os.WIFSTOPPED(status):
                j.stopped[i] = 1
                if j.nrrunning == 0: j.reportstopped()
            else:
                del children[pid]
                j.termed[i] = 1
                j.nrtermed += 1
                if os.WIFSIGNALED(status):
                    sys.stderr.write("[%d]: Terminated by signal %d:%s.\n" % 
                                (pid,os.WTERMSIG(status),
                                 signal_table.keys()[os.WTERMSIG(status)]))
                if j.nrtermed == len(j.pids): 
                    j.reportterminated()
                    j.cleanup()
            if canblock: break    # leave after first info in blocking case
        else: break       # leave when no information is there
    
def spout(st,i):
    '''Skips whitespace by counting up i to the next non-whitespace 
    character.'''
    while i < len(st) and st[i] in string.whitespace: i += 1
    return i

def replacevar(st,i):
    '''Replaces variable. Returns tuple (st2,i2) where st might be another
    string and i2 >= i.'''
    _varprog = re.compile(r'^\$(\w+|\{[^}]*\})')
    m = _varprog.match(st[i:])
    if not m: return (st,i+1)
    b,e = m.span(0)
    name = m.group(1)
    if name[0] == '{' and name[-1] == '}': name = name[1:-1]
    if os.environ.has_key(name):
        return (st[:b+i]+os.environ[name]+st[e+i:],i)
    else:
        return (st[:b+i]+st[e+i:],i)

def replaceuser(st,i):
    '''Replaces tilde. Returns tuple (st2,i2) where st2 might be another 
    string and i2 >= i.'''
    pos = st.find(' ',i)
    if pos >= 0:
        st2 = os.path.expanduser(st[i:pos])+st[pos:]
    else:
        st2 = os.path.expanduser(st[i:])
    if st2[0] == '~':
        return (st[:i]+st2[1:],i)
    else:
        return (st[:i]+st2,i)

bslashtrans = string.maketrans("abfnrtv'","\a\b\f\n\r\t\v'")

def sh(line,inp = None,out = None,toclose = []):
    '''Returns 1 if an "exit" command was issued and 0 otherwise. An exception
    occurs if something goes wrong. Additional errors are printed on stderr.
    inp and out can be file descriptors which will be the default choice
    for input and output redirection respectively for all processes.
    At the end of this function inp and out are closed.
    First of all is to resolve backquotes, thereby not implementing nested
    backquotes or strings containing backquotes within backquotes!'''
    if line.find('\n'):
        lines = line.split('\n')
        line = string.join(lines," ; ")
    posbeg = line.find('`')
    while posbeg >= 0:
        posend = line.find('`',posbeg+1)
        if posend >= 0:
            subline = line[posbeg+1:posend]
            # now we first execute this in the background:
            (pr,pw) = os.pipe()
            subthread = threading.Thread(target=sh,args=(subline,None,pw,[pr]))
            subthread.run()
            buffer = ''
            chunk = os.read(pr,1024)
            while chunk != '':
                buffer += chunk.replace('\n',' ')
                chunk = os.read(pr,1024)
            os.close(pr)
            line = line[:posbeg] + buffer + line[posend+1:]
            posbeg = line.find('`')
        else:
            posbeg = -1
        if posbeg == -1 and \
           not(os.environ.has_key('PYSH_SILENT')): print "Doing:",line
              
    # Input/Output/Error redirection:
    def doredirect(j):
        '''Organizes redirection if started in the current collector. 
        j is the current job. An exception is raised if something goes wrong.'''
        if j.redirect == '': return
        mode = j.redirect
        j.redirect = ''
        st = string.join(j.collector,'')
        if mode == '>':
            try: 
                j.output = os.open(st,os.O_WRONLY|os.O_TRUNC|os.O_CREAT)
            except:
                m = "Error: Cannot open file for output redirection: "+st+"\n"
                sys.stderr.write(m)
                raise ShError,m
        elif mode == '>>':
            try:
                j.output = os.open(st,os.O_WRONLY|os.O_APPEND|os.O_CREAT)
            except:
                m = "Error: Cannot append to file for output redirection: "+\
                    st+"\n"
                sys.stderr.write(m)
                raise ShError,m
        elif mode == '<':
            try:
                j.input = os.open(st,os.O_RDONLY)
            except:
                m = "Error: Cannot open file for input redirection: "+st+"\n"
                sys.stderr.write(m)
                raise ShError,m
        elif mode == '2>':
            if st == '': 
                j.error[-1] = 1
                return 0
            try:
                j.error[-1] = os.open(st,os.O_WRONLY|os.O_TRUNC|os.O_CREAT)
            except:
                m = "Error: Cannot open file for error redirection: "+st+"\n"
                sys.stderr.write(m)
                raise ShError,m
        elif mode == '2>>':
            try:
                j.error[-1] = os.open(st,os.O_WRONLY|os.O_APPEND|os.O_CREAT)
            except:
                m = "Error: Cannot open file for error redirection: "+st+"\n"
                sys.stderr.write(m)
                raise ShError,m
        j.collector = []
        return 0

    # to avoid infinite recursion in variable replacement:
    rectrap = 0
    i = spout(line,0)
    j = job()         # a new job
    if inp != None:
        j.input = os.dup(inp)
        j.toclose.append(inp)   # let it be closed in children!
    if out != None:
        j.output = os.dup(out)
        j.toclose.append(out)   # let it be closed in children!
    j.toclose.extend(toclose)
    mode = 'normal'
    while i < len(line):
        c = line[i]
        if mode == 'normal':
            if c in string.whitespace:
                doredirect(j)
                j.appendarg()
                i = spout(line,i)
            elif c == '|':
                doredirect(j)
                j.addprocess()
                i = spout(line,i+1)
            elif c == '&':
                doredirect(j)
                j.appendarg()
                j.background = 1
                try:
                    j.launch()
                except SystemExit:
                    if inp != None: os.close(inp)
                    if out != None: os.close(out)
                    return 1
                j = job()
                if inp != None:
                    j.input = os.dup(inp)
                    j.toclose.append(inp)   # let it be closed in children!
                if out != None:
                    j.input = os.dup(out)
                    j.toclose.append(out)   # let it be closed in children!
                j.toclose.extend(toclose)
                i = spout(line,i+1)
            elif c == ';':
                doredirect(j)
                j.appendarg()
                j.background = 0
                try:
                    j.launch()
                except SystemExit:
                    if inp != None: os.close(inp)
                    if out != None: os.close(out)
                    return 1
                j = job()
                if inp != None:
                    j.input = os.dup(inp)
                    j.toclose.append(inp)   # let it be closed in children!
                if out != None:
                    j.input = os.dup(out)
                    j.toclose.append(out)   # let it be closed in children!
                j.toclose.extend(toclose)
                i = spout(line,i+1)
            elif c == '"':
                mode = 'dquote'
                i += 1
            elif c == "'":
                mode = 'squote'
                i += 1
            elif c == '\\':
                i += 1
                if i < len(line):
                    j.collector.append(string.translate(line[i],bslashtrans))
                    i += 1
            elif c == '$':
                (line,ii) = replacevar(line,i)   # replaces variable
                if ii == i: 
                    rectrap += 1
                    if rectrap > 1000:
                        pass   # handle recursion trap
                i = ii
            elif c == '~':
                if len(j.collector) == 0:
                    (line,ii) = replaceuser(line,i)
                    if ii == i:
                        rectrap += 1
                        if rectrap > 1000:
                            pass   # handle recursion trap
                    i = ii
                else:
                    j.collector.append(c)
                    i += 1
            elif c == '>':
                if len(j.collector) == 0:
                    if j.output != None: os.close(j.output)
                    if line[i+1:i+2] == '>':
                        j.redirect = '>>'
                        i = spout(line,i+2)
                    else:
                        j.redirect = '>'
                        i = spout(line,i+1)
                else:
                    j.collector.append(c)
                    i += 1
            elif c == '<':
                if len(j.collector) == 0:
                    if j.input != None: os.close(j.input)
                    j.redirect = '<'
                    i = spout(line,i+1)
                else:
                    j.collector.append(c)
                    i += 1
            elif c == '!':
                next = line[i+1:i+2]
                if len(history) > 1:
                    lastline = history[-2].split(" ")
                else:
                    lastline = [""]
                if next == '$':
                    line = line[:i]+lastline[-1]+line[i+2:]
                    if line[i:i+1] == "!": i += 1
                elif next >= '0' and next <= '9' and \
                     len(lastline) > ord(next)-ord('0'):
                    line = line[:i]+lastline[ord(next)-ord('0')]+line[i+2:]
                    if line[i:i+1] == "!": i += 1
                else:
                    j.collector.append(c)
                    i += 1
            elif c == '2':
                if len(j.collector) == 0 and line[i+1:i+2] == '>':
                    if line[i+2:i+3] == '>':
                        j.redirect = '2>>'
                        i = spout(line,i+3)
                    else:
                        j.redirect = '2>'
                        i = spout(line,i+2)
                else:
                    j.collector.append(c)
                    i += 1
            else:
                if c in '*?[]{}':
                    j.collglob = 1   # we will have to do globbing
                j.collector.append(c)
                i += 1
        elif mode == 'squote':
            if c == "'":
                mode = 'normal'
                i += 1
            else:
                j.collector.append(c)
                i += 1
        elif mode == 'dquote':
            if c == '"':
                mode = 'normal'
                i += 1
            elif c == '\\':
                i += 1
                if i < len(line):
                    j.collector.append(string.translate(line[i],bslashtrans))
                    i += 1
            elif c == '$':
                (line,ii) = replacevar(line,i)   # replaces variable
                if ii == i: 
                    rectrap += 1
                    if rectrap > 1000:
                        pass   # handle recursion trap
                i = ii
            else:
                j.collector.append(c)
                i += 1
    doredirect(j)
    j.appendarg()
    retval = 0
    try:
        j.launch()
    except SystemExit:
        retval = 1
    if inp != None: os.close(inp)
    if out != None: os.close(out)
    return retval

def execpyshrc():
    pyshrc = os.path.join(homedir,'.pyshrc')
    if os.access(pyshrc,os.X_OK):
        execfile(pyshrc)

def felixtaste(p):
    '''Shortens a path according to Felix' taste.'''
    l = p.split('/')
    if len(l) <= 2: return p
    else: return l[0]+'/.../'+l[-1]

def expandprompt(s):
    r = []
    i = 0
    while i < len(s):
        if s[i] != '%': r.append(s[i])
        elif i+1 < len(s):
            c = s[i+1]
            if   c == '%': t = '%'
            elif c == '/': t = os.environ.get('PWD','')
            elif c == '~': t = unexpandtilde(os.environ.get('PWD',''))
            elif c == 'f': 
                t = unexpandtilde(os.environ.get('PWD',''))
                t = felixtaste(t)
            elif c == 'M': t = socket.gethostbyaddr(socket.gethostname())[0]
            elif c == 'm': t = socket.gethostname()
            elif c == 't': t = time.strftime('%I:%M%p',time.localtime())
            elif c == 'T': t = time.strftime('%H:%M',time.localtime())
            elif c == '*': t = time.strftime('%H:%M.%S',time.localtime())
            elif c == 'n': t = os.environ.get('USER','')
            elif c == 'w': t = time.strftime('%a %d',time.localtime())
            elif c == 'W': t = time.strftime('%m/%d/%y',time.localtime())
            elif c == 'D': t = time.strftime('%y-%m-%d',time.localtime())
            elif c == 'x': t = time.strftime('%x',time.localtime())
            elif c == '?': 
                if lastexitcode == None:
                    t = 'None'
                else:
                    t = str(lastexitcode >> 8)
            elif c == '#': 
              if os.geteuid() == 0: t = '#'
              else:                 t = '%'
            elif c == 'c': t = os.path.basename(os.environ.get('PWD',''))
            else: t = ''
            r.append(t)
            i += 1
        else:
            r.append('%')
        i += 1
    return string.join(r,'')

def makeprompt():
    if not os.environ.has_key('PROMPT'):
        return '% '
    return expandprompt(os.environ['PROMPT'])

def makerprompt():
    if not(os.environ.has_key('RPROMPT')):
        return ''
    if not(os.environ.has_key('COLUMNS')):
        cols = 80
    else:
        try:
            cols = int(os.environ['COLUMNS'])
        except:
            cols = 80
    rprompt = expandprompt(os.environ['RPROMPT'])
    return ' '*(cols-1-len(rprompt))+rprompt+'\r'


